@@ -74,7 +74,7 @@
   #:use-module ((nyacc lex) #:select (cnumstr->scm))
   #:use-module ((nyacc util) #:select (ugly-print))
   #:re-export (*nyacc-version*)
-  #:version (2 01 6))
+  #:version (2 02 1))
 
 ;; maybe change to a record-type
 (define *options* (make-parameter '()))
@@ -1357,7 +1357,8 @@
 	    (list (string->symbol
 		   (string-append (m-path->name mod) "-symbol-val")) 'k))
 	  ext-mods)))
-    (ppscm `(define ,st-name '(,@defs)))
+    (ppscm `(define-public ,st-name '(,@defs)))
+    (sfscm "(export ~A)\n" st-name)
     (sfscm "\n")
     (ppscm `(define ,sv-name (lambda (k) (or (assq-ref ,st-name k) ,@ext-ftns))))
     (sfscm "(export ~A)\n" sv-name)
@@ -1802,7 +1803,7 @@
 	(fherr "ERROR: not found: ~S" filename))
     (call-with-input-file filename
       (lambda (iport)
-	(let ((output (or (assq-ref 'output options)
+	(let ((output (or (assq-ref options 'output)
 			  (string-append (dirname filename) "/"
 					 (basename filename ".ffi") ".scm"))))
 	  (call-with-output-file/atomic output
