@@ -63,7 +63,9 @@
   #:use-module ((nyacc lex) #:select (cnumstr->scm))
   #:use-module ((nyacc util) #:select (ugly-print))
   #:re-export (*nyacc-version*)
-  #:version (2 01 6))
+  #:version (2 02 1))
+
+(define default-renamer (lambda (name ctxt) name))
 
 ;; maybe change to a record-type
 (define *options* (make-parameter '()))
@@ -72,7 +74,7 @@
 (define *echo-decls* (make-parameter #f)) ; add echo-decls code for debugging
 
 (define *prefix* (make-parameter "")) ; name prefix (e.g., prefix-syms)
-(define *renamer* (make-parameter (lambda (n c) n))) ; renamer from ffi-module
+(define *renamer* (make-parameter default-renamer)) ; renamer from ffi-module
 
 (define *udict* (make-parameter '()))	   ; udecl dict
 (define *ddict* (make-parameter '()))	   ; cpp-def based dict
@@ -223,6 +225,17 @@
    '()
    (pkg-config name "--libs")))
 
+(define (pkg-config-dirs name)
+  (fold-right
+   (lambda (s l)
+     (cond
+      ((< (string-length s) 3) l)
+      ((string=? "-L" (substring/shared s 0 2))
+       (cons (substring/shared s 2) l))
+      (else l)))
+   '()
+   (pkg-config name "--libs")))
+
 (define (resolve-attr-val val)
   (let* ((val (if (procedure? val) (val) val)))
     (cond
@@ -396,9 +409,9 @@
     (match mtail
       (`((pointer-to) (typename ,name))
        (let ((name (rename name 'type)))
-	 (if (dmem? (w/* name) defined)
-	     (strings->symbol name "*")
-             `(cpointer ,(mtail->ctype (cdr mtail))))))
+         (if (dmem? (w/* name) defined)
+             (strings->symbol name "*")
+             `(cpointer (delay ,(string->symbol name))))))
       (`((pointer-to) (void))
        `(cpointer 'void))
       (`((pointer-to) (fixed-type "char"))
@@ -413,12 +426,18 @@
                         (function*-wraps return params)))
          (lambda (pc->pr pr->pc)
            `(cpointer (cfunction ,pc->pr ,pr->pc)))))
-      (`((pointer-to) (pointer-to) (function-returning . ,rest) . ,rest)
+      (`((pointer-to) (pointer-to) (function-returning . ,_1) . ,_2)
        `(cpointer (cbase 'void*)))
       (`((pointer-to) (struct-ref (ident ,name)))
-       (if (dmem? (w/struct name) defined)
-           `(cpointer ,(strings->symbol "struct-" name))
-	   `(cpointer 'void)))
+       (let* ((name (rename name 'type)) (aggr-name (sfsym "struct-~a" name)))
+         (cond
+          ((dmem? (w/struct name) defined) `(cpointer aggr-name))
+	  (else `(cpointer (delay ,aggr-name)))))) ;; was (cpointer 'void)
+      (`((pointer-to) (union-ref (ident ,name)))
+       (let* ((name (rename name 'type)) (aggr-name (sfsym "union-~a" name)))
+         (cond
+          ((dmem? (w/struct name) defined) `(cpointer aggr-name))
+	  (else `(cpointer (delay ,aggr-name)))))) ;; was (cpointer 'void)
       (`((pointer-to) . ,rest)
        `(cpointer ,(mtail->ctype rest)))
       (`((array-of ,n) . ,rest)
@@ -490,7 +509,9 @@
 	 (pkg-config (assq-ref attrs 'pkg-config))
          (libs (resolve-attr-val (assq-ref attrs 'library)))
          (libs (if pkg-config (append (pkg-config-libs pkg-config) libs) libs))
- 	 (libs (delete "libm" libs)))
+ 	 (libs (delete "libm" libs))
+         (dirs (resolve-attr-val (assq-ref attrs 'lib-dirs)))
+         (dirs (if pkg-config (append (pkg-config-dirs pkg-config) dirs) dirs)))
     (sfscm ";; generated with `guild compile-ffi ~A.ffi'\n"
 	   (m-path->f-path path))
     (nlscm)
@@ -510,10 +531,12 @@
     (sfscm "\n")
     (ppscm
      `(define (foreign-pointer-search name)
+        (define (flc l)
+          (load-foreign-library (car l) #:search-path (list ,@dirs)))
         (let loop ((libs (list #f ,@libs)))
           (cond
            ((null? libs) (error "no library for ~s" name))
-           ((false-if-exception (foreign-library-pointer (car libs) name)))
+           ((false-if-exception (foreign-library-pointer (flc libs) name)))
            (else (loop (cdr libs)))))))
     (sfscm "\n")
     (if (*echo-decls*) (sfscm "(define echo-decls #t)\n\n"))))
@@ -533,19 +556,19 @@
 
   (define (cnvt mtail)
     (match mtail
-      (`((pointer-to) . ,rest) ''*)
+      (`((pointer-to) . ,_1) ''*)
       (`((fixed-type ,name))
        (or (assoc-ref ffi-typemap name)
 	   (fherr/once "no FFI type for ~A" name)))
       (`((float-type ,name))
        (or (assoc-ref ffi-typemap name)
 	   (fherr/once "no FFI type for ~S" name)))
-      (`((typename ,name) . ,rest)
+      (`((typename ,name) . ,_1)
        (or (assoc-ref ffi-typemap name)
 	   (fherr "no FFI type for ~S" name)))
       (`((void)) 'ffi:void)
-      (`((enum-def . ,rest2) . ,rest1) 'ffi:int)
-      (`((enum-ref . ,rest2) . ,rest1) 'ffi:int)
+      (`((enum-def . ,_1) . ,_2) 'ffi:int)
+      (`((enum-ref . ,_1) . ,_2) 'ffi:int)
 
       (`((array-of ,dim) . ,rest)
        `(make-list ,(eval-dim dim) ,(cnvt rest)))
@@ -584,14 +607,15 @@
   (let* ((udecl (expand-typerefs
                  `(udecl (decl-spec-list (type-spec (typename ,name)))
                          (init-declr (ident ,mname))) (*udict*)
-                         '((enum . "*any*")))) ;; hack provided 
+                         '((enum . "*any*")))) ;; hack provided
          (mdecl (udecl->mdecl udecl)))
     (match (md-tail mdecl)
-      (`((pointer-to) . ,_0) `(unwrap-pointer ,mname ,(string->symbol name)))
-      (`((fixed ,name)) `(unwrap-number ,mname))
-      (`((float ,name)) `(unwrap-number ,mname))
-      (`((enum-def . ,_1)) `(unwrap~enum  ,mname))
-      (`((enum-ref . ,_1)) `(unwrap~enum  ,mname))
+      (`((pointer-to) . ,_1)
+       `(cdata-arg->pointer ,mname ,(string->symbol name)))
+      (`((fixed ,name)) `(cdata-arg->number ,mname))
+      (`((float ,name)) `(cdata-arg->number ,mname))
+      (`((enum-def . ,_1)) `(unwrap~enum ,mname))
+      (`((enum-ref . ,_1)) `(unwrap~enum ,mname))
       (__ #f))))
 
 (define (unwrap-mdecl mdecl)
@@ -599,33 +623,34 @@
         (mname (string->symbol (md-label mdecl)))
         (mtail (md-tail mdecl)))
     (match (car mtail)
-      (`(fixed-type ,name) `(unwrap-number ,mname))
-      (`(float-type ,name) `(unwrap-number ,mname))
+      (`(fixed-type ,name) `(cdata-arg->number ,mname))
+      (`(float-type ,name) `(cdata-arg->number ,mname))
       (`(typename ,name)
        (cond
-	((dmem? name def-defined) `(unwrap-number ,mname))
+	((dmem? name def-defined) `(cdata-arg->number ,mname))
 	((dmem? name defined) (defined-type-unwrapper name mname))
 	(else #f)))
       (`(pointer-to)
        (match mtail
          (`((pointer-to) (typename ,name))
           (cond
-           ((dmem? name def-defined) `(unwrap-pointer ,mname))
+           ((dmem? name def-defined) `(cdata-arg->pointer ,mname))
            ((dmem? (w/* name) defined)
-            `(unwrap-pointer ,mname ,(strings->symbol name "*")))
+            `(cdata-arg->pointer ,mname ,(strings->symbol name "*")))
            ((dmem? name defined)
-            `(unwrap-pointer ,mname (cpointer ,(string->symbol name))))
-           (else `(unwrap-pointer ,mname))))
+            `(cdata-arg->pointer ,mname (cpointer ,(string->symbol name))))
+           (else `(cdata-arg->pointer ,mname))))
          (`((pointer-to) (function-returning (param-list . ,params)) . ,rest)
           (let ((return (mdecl->udecl (cons "~ret" rest))))
             (call-with-values
                 (lambda () (function*-wraps return params))
               (lambda (pr pc)
                 (let ((pc '(lambda (p) 'unused))) ; only proc->ptr is used
-                  `(unwrap-pointer ,mname (cpointer (cfunction ,pr ,pc))))))))
+                  `(cdata-arg->pointer
+                    ,mname (cpointer (cfunction ,pr ,pc))))))))
          (`((pointer-to) . ,_1)
-          `(unwrap-pointer ,mname))))
-      (`(function-returning . ,_0)
+          `(cdata-arg->pointer ,mname))))
+      (`(function-returning . _0)
        (unwrap-mdecl (cons (md-label mdecl) (cons '(pointer-to) mtail))))
       (`(enum-def (ident ,name) ,_1)
        (cond
@@ -640,8 +665,8 @@
 	(else `(unwrap~enum ,mname))))
       (`(struct-ref (ident ,name)) `(cdata&-ref ,mname))
       (`(union-ref (ident ,name)) `(cdata&-ref ,mname))
-      (`(array-of ,size) `(unwrap-array ,mname))
-      (`(array-of) `(unwrap-array ,mname))
+      (`(array-of ,size) `(cdata-arg->pointer ,mname))
+      (`(array-of) `(cdata-arg->pointer ,mname))
       ;; not expected
       (`(void) #f)
       (`(struct-def . ,_1) `(cdata&-ref ,mname))
@@ -653,7 +678,7 @@
   (let* ((udecl (expand-typerefs
                  `(udecl (decl-spec-list (type-spec (typename ,name)))
                          (init-declr (ident ,mname))) (*udict*)
-                         '((enum . "*any*")))) ;; hack provided 
+                         '((enum . "*any*")))) ;; hack provided
          (mdecl (udecl->mdecl udecl)))
     (match (md-tail mdecl)
       (`((enum-def . ,_1)) (list (sfsym "wrap-~a" name) mname))
@@ -864,7 +889,9 @@
             ((type-qual)
              (loop ss (cons (sx-tag (sx-ref (car tl) 1)) tq) ts (cdr tl)))
             ((type-spec)
-             (loop ss tq (sx-ref (car tl) 1) (cdr tl)))))))
+             (loop ss tq (sx-ref (car tl) 1) (cdr tl)))
+            ((fctn-spec)
+             (loop ss tq ts (cdr tl)))))))
 
   (define ftn-declr?
     (let ((ff (node-closure
@@ -879,11 +906,11 @@
   (define (bkref-getall attr)
     (and=> (assq-ref attr 'typedef) (lambda (t) (string-split (car t) #\,))))
 
-  (*defined* defined)                   ; set global for converters
-
   (define-syntax-rule (deftype name type)
     `(define-public ,name (name-ctype ',name ,type)))
 
+  (*defined* defined)                 ; set global for converters
+
   (let* ((tag attr specl declr (split-udecl udecl))
          (sspec tqual tspec (specl-props specl)))
 
@@ -927,6 +954,23 @@
                       `(define ,*type (cfunction ,pc->pr ,pr->pc))
                       (deftype type `(cpointer ,*type)))))))))
 
+          ;;; CHECK
+          (`((pointer-to) (struct-ref (ident ,aggr)))
+           (values
+            (dcons name defined)
+            (let ((aggr-name (sfsym "struct-~a" aggr)))
+              (xcons* seed
+                `(define-public ,aggr-name 'void)
+                (deftype type `(cpointer (delay ,aggr-name)))))))
+
+          (`((pointer-to) (union-ref (ident ,aggr)))
+           (values
+            (dcons name defined)
+            (let ((aggr-name (sfsym "union-~a" aggr)))
+              (xcons* seed
+                `(define-public ,aggr-name 'void)
+                (deftype type `(cpointer (delay ,aggr-name)))))))
+
           (`((pointer-to) . ,rest)
            (values
             (case (caar rest)
@@ -937,7 +981,7 @@
 
           (__
            (sx-match (car mtail)
-             
+
              ((struct-def (@ . ,attr) (ident ,agname) ,field-list)
               (let ((agname (rename agname 'type)))
                 (values
@@ -1017,7 +1061,8 @@
                      (deftype type* `(cpointer 'void))))))))
 
              (((fixed-type float-type) ,basename)
-              (values defined (cons (deftype type (mtail->ctype mtail)) seed)))
+              (values (dcons name defined)
+                      (cons (deftype type (mtail->ctype mtail)) seed)))
 
              ((enum-def ,enum-def-list)
               (values
@@ -1081,10 +1126,10 @@
                            (dcons (w/* name) defined)
                            (cons (deftype type* atype*) seed)))
                         (values defined seed))))
-	        (else
-	         (let ((xdecl (expand-typerefs udecl (*udict*) defined)))
-	           (udecl->sexp xdecl udict defined seed))))))
-             
+	         (else
+	          (let ((xdecl (expand-typerefs udecl udict defined)))
+	            (udecl->sexp xdecl udict defined seed))))))
+
              (,__
               (sferr "udecl->sexp missed typedef:\n") (pperr mdecl)
               (values defined seed)))))))
@@ -1167,7 +1212,7 @@
 	    (else
              (let* ((type (sfsym "enum-~a" enum-name))
                     (defs (canize-enum-def-list
-                           enum-def-list (*udict*) (*ddict*)))
+                           enum-def-list udict (*ddict*)))
                     (enums (enum-def-list->alist defs)))
 	       (values
                 (dcons (w/enum enum-name) defined)
@@ -1191,8 +1236,10 @@
              (name (md-label mdecl))
              (rname (rename name 'variable))
              (mtail (cdr (md-tail mdecl))) ; remove (extern)
-             (mtail* `((pointer-to) . ,mtail))
-             (type* (mtail->ctype mtail*))
+             ;;(mtail* `((pointer-to) . ,mtail))
+             ;;(type* (mtail->ctype mtail*))
+             (ctype (mtail->ctype mtail))
+             (type* `(cpointer ,ctype))
              (name* (strings->symbol name "*")))
         (values
          defined
@@ -1207,7 +1254,7 @@
 	          ((arg) (cdata-set! (force obj) '* arg)))))))))
 
      ((memq 'const sspec)
-      (let ((udecl (expand-typerefs udecl (*udict*) (*defined*)))
+      (let ((udecl (expand-typerefs udecl udict defined))
             (mdecl (udecl->mdecl udecl)))
         (sferr "skipping const expression: can't do this yet"))
       (values defined seed))
@@ -1243,10 +1290,11 @@
 	  (fold
 	   (lambda (name seed)
 	     (let* ((repl (expand-cpp-name name ddict '()))
-		    (val (and (string? repl) (eval-code-string repl))))
+		    (val (and (string? repl)
+                              (false-if-exception (eval-code-string repl)))))
 	       (if val
-		   (acons (string->symbol name) val seed)
-		   seed)))
+                   (acons (string->symbol name) val seed)
+	           seed)))
 	   '() mod-def-names))
 	(ext-ftns			; lookup in use-ffi-modules
 	 (map
@@ -1255,6 +1303,7 @@
 		   (string-append (m-path->name mod) "-symbol-val")) 'k))
 	  ext-mods)))
     (ppscm `(define ,st-name '(,@defs)))
+    (sfscm "(export ~A)\n" st-name)
     (sfscm "\n")
     (ppscm `(define ,sv-name (lambda (k) (or (assq-ref ,st-name k) ,@ext-ftns))))
     (sfscm "(export ~A)\n" sv-name)
@@ -1474,7 +1523,7 @@
 ;; Convert a snippet of C code to list of scheme forms.
 ;; @example
 ;; > (ccode->sexp "double sqrt(doubld);")
-;; $1 = 
+;; $1 =
 ;; @end example
 ;; @end deffn
 (define* (ccode->sexp code #:optional attrs)
@@ -1515,7 +1564,7 @@
           (use-modules (nyacc lang c99 ffi-help))
           (load-include-file \"cairo.h\" #:pkg-config \"cairo\")"
   (parameterize ((*options* '()) (*defined* '())
-		 (*renamer* (lambda (n c) n)) (*errmsgs* '())
+		 (*renamer* default-renamer) (*errmsgs* '())
 		 (*prefix* "") (*mport* #t) (*udict* '()))
     (let* ((attrs (acons 'include (list filename) '()))
 	   (attrs (if pkg-config (acons 'pkg-config pkg-config attrs) attrs))
@@ -1534,10 +1583,13 @@
       (ppscm '(use-modules (nyacc foreign cdata)))
       (ppscm
        `(define (foreign-pointer-search name)
+          (define (flc l)
+            (load-foreign-library
+             (car l) #:search-path (list ,@(pkg-config-dirs pkg-config))))
           (let loop ((libs (list #f ,@(pkg-config-libs pkg-config))))
             (cond
              ((null? libs) (fherr "not found: ~s" name))
-             ((false-if-exception (foreign-library-pointer (car libs) name)))
+             ((false-if-exception (foreign-library-pointer (flc libs) name)))
              (else (loop (cdr libs)))))))
       (process-decls decls udict def-defined)
       (close (*mport*))
@@ -1587,7 +1639,7 @@
 	   (member key '(#:api-code
 			 #:cpp-defs #:decl-filter #:inc-dirs #:inc-filter
 			 #:inc-help #:include #:library #:pkg-config #:renamer
-			 #:use-ffi-module #:def-keepers))))
+			 #:use-ffi-module #:def-keepers #:lib-dirs))))
     (define (module-option? key) (keyword? key))
 
     (syntax-case x ()
@@ -1627,13 +1679,13 @@
 ;; @end deffn
 (define* (compile-ffi-file filename #:optional (options '()))
   (parameterize ((*options* options) (*defined* '())
-		 (*renamer* (lambda (n c) n)) (*errmsgs* '()) (*prefix* "")
+		 (*renamer* default-renamer) (*errmsgs* '()) (*prefix* "")
 		 (*mport* #t) (*udict* '()) (*ddict* '()))
     (if (not (access? filename R_OK))
 	(fherr "ERROR: not found: ~S" filename))
     (call-with-input-file filename
       (lambda (iport)
-	(let ((output (or (assq-ref 'output options)
+	(let ((output (or (assq-ref options 'output)
 			  (string-append (dirname filename) "/"
 					 (basename filename ".ffi") ".scm"))))
 	  (call-with-output-file/atomic output
